"""
Fields for geo data

See: http://docs.mongodb.org/manual/applications/geospatial-indexes/

GeoJSON: http://geojson.org/geojson-spec.html

"""
from collections.abc import Sequence

from yadm.documents import DocumentItemMixin
from yadm.fields.base import Field


TYPES = []


def _geo_type(type):
    # class decorator for add geo types to TYPES
    TYPES.append(type)
    return type


class Geo(DocumentItemMixin):
    """ Base class for GeoJSON data
    """
    type = None


class GeoCoordinates(Geo):
    """ Base class for GeoJSON data with coordinates
    """
    def get_coordinates(self):
        raise NotImplementedError('get_coordinates must be implemented')

    def to_mongo(self):
        return {
            'type': self.type,
            'coordinates': self.get_coordinates(),
        }


@_geo_type
class Point(GeoCoordinates):
    """ Class for GeoJSON Point objects

    See: http://geojson.org/geojson-spec.html#id2
    """
    type = 'Point'

    def __init__(self, longitude, latitude):
        self.longitude = longitude
        self.latitude = latitude

    def __getitem__(self, idx):
        return (self.longitude, self.latitude)[idx]

    def get_coordinates(self):
        return [self.longitude, self.latitude]

    @classmethod
    def from_mongo(cls, data):
        try:
            coordinates = data['coordinates']
        except KeyError:
            raise ValueError('coordinates not found in data: "{!r}"'.format(data))

        try:
            longitude, latitude = coordinates
        except IndexError:
            raise ValueError('wrong coordinates in data: "{!r}"'.format(data))

        return cls(longitude, latitude)


@_geo_type
class MultiPoint(GeoCoordinates, Sequence):
    """ Class for GeoJSON MultiPoint objects

    See: http://geojson.org/geojson-spec.html#id5
    """
    type = 'MultiPoint'

    def __init__(self, points):
        self._points = points

    def __len__(self):
        return len(self._points)

    def __iter__(self):
        return self._points

    def __getitem__(self, item):
        return self._points[item]

    def get_coordinates(self):
        return [p.to_mongo()['coordinates'] for p in self._points]

    @classmethod
    def from_mongo(cls, data):
        try:
            coordinates = data['coordinates']
        except KeyError:
            raise ValueError('coordinates not found in data: "{!r}"'.format(data))

        return cls([Point(*c) for c in coordinates])


class GeoField(Field):
    """ Base field for GeoJSON objects
    """
    def __init__(self, types=TYPES):
        self.types = types
        self.types_dict = {t.type: t for t in types}

    def to_mongo(self, document, geo):
        return geo.to_mongo() if geo is not None else None

    def from_mongo(self, document, data):
        if data is None or isinstance(data, Geo):
            return data
        elif data is None:
            return None

        geo_type = self.types_dict.get(data['type'])

        if geo_type is None:
            raise ValueError('unknown type in data: "{!r}"'.format(data))

        return geo_type.from_mongo(data)


class GeoOneTypeField(GeoField):
    """ Base field for GeoJSON objects
        with one acceptable type
    """
    type = None

    def __init__(self):
        if self.type is None:
            raise NotImplementedError('attribute "type" must be implemented')

        self.types = [self.type]
        self.types_dict = {self.type.type: self.type}

    def prepare_value(self, document, value):
        if isinstance(value, self.type) or value is None:
            return value
        elif isinstance(value, dict):
            return self.type.from_mongo(value)


class PointField(GeoOneTypeField):
    """ Field for Point
    """
    type = Point


class MultiPointField(GeoOneTypeField):
    """ Field for MultiPoint
    """
    type = MultiPoint



"""
This package contain all fields.
"""

from yadm.fields.base import FieldDescriptor, Field

from yadm.fields.simple import (
    ObjectIdField,
    BooleanField,
    IntegerField,
    FloatField,
    StringField,
)

from yadm.fields.email import EmailField

from yadm.fields.datetime import DatetimeField
from yadm.fields.decimal import DecimalField
from yadm.fields.money import Money, MoneyField

from yadm.fields.embedded import EmbeddedDocumentField

from yadm.fields.list import ListField
from yadm.fields.set import SetField
from yadm.fields.map import MapField
from yadm.fields.map import MapIntKeysField, MapObjectIdKeysField

from yadm.fields.reference import ReferenceField

from yadm.fields.geo import Point, PointField
from yadm.fields.geo import MultiPoint, MultiPointField



from collections.abc import Sequence

from pymongo.errors import BulkWriteError

from yadm.serialize import to_mongo


class Bulk:
    """ Bulk object

    :param Database db: Database instance
    :param MetaDocument document_class: document class for collection
    :param bool ordered: create ordered bulk (default `False`)
    :param bool raise_on_errors: raise BulkWriteError exception
        if write errors (default `True`)

    Context manager:

        with db.bulk(Doc) as bulk:
            bulk.insert(doc_1)
            bulk.insert(doc_2)
    """
    result = None
    error = False

    def __init__(self, db, document_class,
                 ordered=False, raise_on_errors=True):
        self._db = db
        self._document_class = document_class
        self._ordered = ordered
        self._raise_on_errors = raise_on_errors

        self._collection = db._get_collection(self._document_class)

        if ordered:
            self._bulk_mongo = self._collection.initialize_ordered_bulk_op()
        else:
            self._bulk_mongo = self._collection.initialize_unordered_bulk_op()

    def __repr__(self):
        return "<{}.{}('{}') at {}>".format(
            self.__class__.__module__,
            self.__class__.__name__,
            self._collection.name,
            hex(id(self)),
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            self.execute()

    def execute(self):
        try:
            raw_data = self._bulk_mongo.execute()
        except BulkWriteError as exc:
            self.error = True
            raw_data = exc.details

        self.result = BulkResult(self, raw_data)

        if self.error and self._raise_on_errors:
            raise BulkWriteError(raw_data)

        return self.result

    def insert(self, document):
        """ Add insert document to bulk

        :param Document document: document for insert

        .. warning::
            This unlike `Database.insert`!
            It not set `document.id` and `document.__db__`.
        """
        if not isinstance(document, self._document_class):
            raise TypeError("Bulk.insert() argument must be a {}, not '{}'"
                            "".format(self._document_class, document.__class__))

        self._bulk_mongo.insert(to_mongo(document))


class BulkResult:
    """ Object who provide result of `Bulk.execute()`
    """
    def __init__(self, bulk, raw_data):
        self._bulk = bulk
        self._raw_data = raw_data

        self.n_inserted = self._raw_data['nInserted']
        self.n_upserted = self._raw_data['nUpserted']
        self.n_modified = self._raw_data['nModified']
        self.n_removed = self._raw_data['nRemoved']

    def __bool__(self):
        return not self._bulk.error

    @property
    def write_errors(self):
        return _BulkResultWriteErrors(
            self._bulk, self._raw_data['writeErrors'])

    @property
    def write_concern_errors(self):
        # TODO
        return self._raw_data['writeConcernErrors']

    @property
    def upserted(self):
        # TODO
        return self._raw_data['upserted']


class _BulkResultWriteErrors(Sequence):
    def __init__(self, bulk, raw_data):
        self._bulk = bulk
        self._raw_data = raw_data

    def __getitem__(self, idx):
        if isinstance(idx, slice):
            raise TypeError("Slicing not supported")
        else:
            return _BulkResultWriteErrorsItem(self._bulk, self._raw_data[idx])

    def __len__(self):
        return len(self._raw_data)


class _BulkResultWriteErrorsItem:
    _document_cache = None

    def __init__(self, bulk, raw_data):
        self._bulk = bulk
        self._raw_data = raw_data

        self.index = self._raw_data['index']
        self.code = self._raw_data['code']
        self.errmsg = self._raw_data['errmsg']
        self.op = self._raw_data['op']

    @property
    def document(self):
        if self._document_cache is None:
            self._document_cache = self._bulk._document_class(**self.op)

        return self._document_cache




from unittest import TestCase

from yadm import (
    Document,
    EmbeddedDocument,
)
from yadm.common import EnclosedDocDescriptor
from yadm.fields import (
    Field,
    BooleanField,
    IntegerField,
    FloatField,
    StringField,
    EmbeddedDocumentField,
    ReferenceField,
)


class Foo(Document):
    f = BooleanField(True)


class Bar(EmbeddedDocument):
    f = IntegerField(42)


class Baz(Document):
    f = FloatField(3.14)


class Qux(EmbeddedDocument):
    f = StringField('quux')


class Corge(Document):
    grault = EmbeddedDocumentField(Foo)
    garply = ReferenceField(Bar)
    waldo = EmbeddedDocumentField(Baz)
    fred = ReferenceField(Qux)
    plugh = EmbeddedDocumentField('self')
    xyzzy = ReferenceField('self')


class Thud(Document):
    rf = ReferenceField('dumb.path')


class Thud2(EmbeddedDocument):
    ef = EmbeddedDocumentField('another.dumb.path')


class FooDocField(Field):
    reference_document_class = EnclosedDocDescriptor('reference')
    embedded_document_class = EnclosedDocDescriptor('embedded')


class EnclosedDocDescriptorTest(TestCase):

    def setUp(self):
        self.R = ReferenceField(Foo)
        self.E = EmbeddedDocumentField(Bar)
        self.CR = ReferenceField('tests.test_common.Baz')
        self.CE = EmbeddedDocumentField('tests.test_common.Qux')

    def assertReference(self, classinfo, expected_value):
        self.assertEquals(classinfo.reference_document_class, expected_value)
        self.assertEquals(getattr(classinfo, '_reference_document_class'),
                          expected_value)

    def assertEmbedded(self, classinfo, expected_value):
        self.assertEquals(classinfo.embedded_document_class, expected_value)
        self.assertEquals(getattr(classinfo, '_embedded_document_class'),
                          expected_value)

    def test_reference(self):
        self.assertReference(self.R, Foo)

    def test_embedded(self):
        self.assertEmbedded(self.E, Bar)

    def test_resolving(self):
        self.assertReference(self.CR, Baz)
        self.assertEmbedded(self.CE, Qux)

        self.assertEmbedded(Corge.grault, Foo)
        self.assertReference(Corge.garply, Bar)
        self.assertEmbedded(Corge.waldo, Baz)
        self.assertReference(Corge.fred, Qux)
        self.assertEmbedded(Corge.plugh, Corge)
        self.assertReference(Corge.xyzzy, Corge)

        with self.assertRaises(ImportError):
            Thud.rf.reference_document_class
        with self.assertRaises(ImportError):
            Thud2.ef.embedded_document_class

    def test_lowlevel(self):
        # wrong enclosed_cls_type value
        try:
            EnclosedDocDescriptor('reference')
            EnclosedDocDescriptor('embedded')
        except Exception as e:
            self.assertTrue(False, 'Unexpected exception: {!r}'.format(e))
        for cls_type in ('waldo', 'fred'):
            with self.assertRaises(ValueError):
                EnclosedDocDescriptor(cls_type)

        # class binding
        self.assertIsInstance(FooDocField.reference_document_class,
                              EnclosedDocDescriptor)
        self.assertIsInstance(FooDocField.embedded_document_class,
                              EnclosedDocDescriptor)

        # getattr --> None
        doc_field = FooDocField()
        self.assertIsNone(doc_field.reference_document_class)
        self.assertIsNone(doc_field.embedded_document_class)

        doc_field.reference_document_class = Foo
        doc_field.embedded_document_class = Bar

        self.assertReference(doc_field, Foo)
        self.assertEmbedded(doc_field, Bar)

        # destructor
        del doc_field.reference_document_class
        self.assertIsNone(getattr(doc_field, '_reference_document_class', None))
        del doc_field.embedded_document_class
        self.assertIsNone(getattr(doc_field, '_embedded_document_class', None))



from bson import ObjectId

from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class SetReferenceFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestRDoc(Document):
            __collection__ = 'testrdocs'
            i = fields.IntegerField()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            li = fields.SetField(fields.ReferenceField(TestRDoc))

        self.TestRDoc = TestRDoc
        self.TestDoc = TestDoc

    def test_save(self):
        td = self.TestDoc()

        trd_one = self.TestRDoc()
        trd_one.i = 13
        self.db.insert(trd_one)
        td.li.add(trd_one)

        trd_two = self.TestRDoc()
        trd_two.i = 42
        self.db.insert(trd_two)
        td.li.add(trd_two)

        self.db.insert(td)

        data = self.db.db.testdocs.find_one()

        self.assertIn('li', data)
        self.assertEqual(len(data['li']), 2)

        res = set()

        for item in data['li']:
            self.assertIsInstance(item, ObjectId)
            res.add(item)

        self.assertEquals(res, {trd_one.id, trd_two.id})

    def test_load(self):
        self.db.db.testdocs.insert(
            {'li': [
                self.db.db.testrdocs.insert({'i': 13}),
                self.db.db.testrdocs.insert({'i': 42}),
            ]}
        )

        doc = self.db.get_queryset(self.TestDoc).find_one()

        self.assertTrue(hasattr(doc, 'li'))
        self.assertEqual(len(doc.li), 2)

        for item in doc.li:
            self.assertIsInstance(item, self.TestRDoc)

        res = set()

        for item in doc.li:
            self.assertIsInstance(item, self.TestRDoc)
            self.assertTrue(hasattr(item, 'i'))
            self.assertIsInstance(item.i, int)
            res.add(item.i)

        self.assertEquals(res, {13, 42})



"""
Work with references.

.. code-block:: python

    class RDoc(Document):
        i = fields.IntegerField

    class Doc(Document):
        rdoc = fields.ReferenceField(RDoc)

    rdoc = RDoc()
    rdoc.i = 13
    db.insert(rdoc)

    doc = Doc()
    doc.rdoc = rdoc
    db.insert(doc)

    doc = db.get_queryset(Doc).with_id(doc.id)  # reload doc
    assert doc.rdoc.id == rdoc.id
    assert doc.rdoc.i == 13


TODO: many2many collections
"""
from yadm.common import EnclosedDocDescriptor
from yadm.documents import Document
from yadm.markers import NoDefault
from yadm.fields.base import Field, FieldDescriptor
from yadm.serialize import from_mongo


class ReferenceFieldDescriptor(FieldDescriptor):
    """ Descriptor for ReferenceField

    Save document in `document.__fields__` after first get.
    """
    def __get__(self, instance, owner):
        value = super().__get__(instance, owner)

        if instance is not None and isinstance(value, Document):
            instance.__data__[self.name] = value

        return value


class ReferenceField(Field):
    """ Field for work with references

    :param reference_document_class: class for refered documents
    """
    descriptor_class = ReferenceFieldDescriptor
    reference_document_class = EnclosedDocDescriptor('reference')

    def __init__(self, reference_document_class, default=NoDefault):
        self.reference_document_class = reference_document_class
        self.default = default

    def copy(self):
        return self.__class__(self.reference_document_class)

    def prepare_value(self, document, value):
        if isinstance(value, Document):
            return value
        else:
            return self.from_mongo(document, value)

    def from_mongo(self, document, value):
        if value is None:
            return None

        elif isinstance(value, dict):
            return from_mongo(self.reference_document_class, value)

        elif isinstance(value, Document):
            return value

        else:
            if document is not None and document.__db__ is not None:
                qs = document.__db__.get_queryset(self.reference_document_class)
                doc = qs.with_id(value)

                if doc:
                    return doc
                else:
                    field = self.reference_document_class.__fields__['_id']
                    return field.prepare_value(None, value)

            else:
                return value

    def to_mongo(self, document, value):
        if isinstance(value, dict):
            return value['_id']

        elif isinstance(value, Document):
            return value.id

        else:
            return value



from yadm import fields
from yadm.documents import Document

from .test_database import BaseDatabaseTest


class SimpleSetFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            s = fields.SetField(fields.IntegerField)

        self.TestDoc = TestDoc

    def test_default(self):
        td = self.TestDoc()
        self.assertIsInstance(td.s, fields.set.Set)
        self.assertFalse(td.s)
        self.assertEqual(len(td.s), 0)
        self.assertEqual(td.s._data, set())

    def test_get(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertTrue(td.s)
        self.assertEqual(len(td.s), 3)
        self.assertEqual(td.s._data, {1, 2, 3})
        self.assertEqual(set(td.s), {1, 2, 3})
        self.assertRaises(TypeError, td.s.__getitem__, 1)

    def test_add(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.add(4)

        self.assertEqual(td.s, {1, 2, 3, 4})

    def test_add_typeerror(self):
        td = self.TestDoc()
        self.assertRaises(ValueError, td.s.add, 'not a number')

    def test_add_save(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.add(4)
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['s'], [1, 2, 3, 4])

    def test_remove(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.remove(2)

        self.assertEqual(td.s, {1, 3})

    def test_remove_save(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.remove(2)
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['s'], [1, 3])

    def test_add_to_set(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.add_to_set(4)

        self.assertEqual(td.s, {1, 2, 3, 4})

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['s'], [1, 2, 3, 4])

    def test_add_to_set_typeerror(self):
        td = self.TestDoc()
        self.assertRaises(ValueError, td.s.add_to_set, 'not a number')

    def test_pull(self):
        _id = self.db.db.testdoc.insert({'s': [1, 2, 3]})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)
        td.s.pull(2)

        self.assertEqual(td.s, {1, 3})

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data['s'], [1, 3])



"""
This module for provide work with MongoDB database.

.. code-block:: python

    import pymongo
    from yadm.database import Database

    from mydocs import Doc

    client = pymongo.MongoClient("localhost", 27017)
    db = Database(self.client, 'test')

    doc = Doc()
    db.insert(doc)

    doc.arg = 13
    db.save(doc)

    qs = db.get_queryset(Doc).find({'arg': {'$gt': 10}})
    for doc in qs:
        print(doc)

"""
from yadm.queryset import QuerySet
from yadm.bulk import Bulk
from yadm.serialize import to_mongo


class Database:
    """ Main object who provide work with database

    :param pymongo.Client client: database connection
    :param str name: database name
    """
    def __init__(self, client, name):
        self.client = client
        self.name = name
        self.db = client[name]

    def __repr__(self):
        return 'Database({!r})'.format(self.db)

    def __call__(self, *args, **kwargs):
        return self.get_queryset(*args, **kwargs)

    def _get_collection(self, document_class):
        """ Return pymongo collection for document class
        """
        return self.db[document_class.__collection__]

    def get_queryset(self, document_class):
        """ Return queryset for document class

        :param document_class: :class:`yadm.documents.Document`

        This create instance of :class:`yadm.queryset.QuerySet`
        with presetted document's collection information.
        """
        return QuerySet(self, document_class)

    def insert(self, document):
        """ Insert document to database

        :param Document document: document instance for insert to database

        It's set :attr:`yadm.documents.Document._id`.
        """
        document.__db__ = self
        collection = self._get_collection(document.__class__)
        document._id = collection.insert(to_mongo(document))
        document.__fields_changed__.clear()
        return document

    def save(self, document, full=False, upsert=False):
        """ Save document to database

        :param Document document: document instance for save
        :param bool full: fully resave document
            (default: `False`)
        :param bool upsert: see documentation for MongoDB's `update`
            (default: `False`)

        If document has not `id` this :meth:`insert` new document.
        """
        if hasattr(document, '_id'):
            document.__db__ = self

            if full:
                self._get_collection(document).update(
                    {'_id': document.id},
                    to_mongo(document),
                    upsert=upsert,
                    multi=False,
                )
            else:
                self._get_collection(document).update(
                    {'_id': document.id},
                    {'$set': to_mongo(
                        document,
                        exclude=['_id'],
                        include=document.__fields__.keys()),
                        # include=document.__fields_changed__),  # must be!
                    },
                    upsert=upsert,
                    multi=False,
                )

            document.__fields_changed__.clear()
            return document
        else:
            return self.insert(document)

    def remove(self, document):
        """ Remove document from database

        :param Document document: document instance for remove from database
        """
        return self._get_collection(document.__class__).remove(document._id)

    def reload(self, document, new_instance=False):
        """ Reload document

        :param Document document: document for reload
        :param bool new_instance: if `True` return new instance of document,
            else change data in given document (default: `False`)
        """
        new = self.get_queryset(document.__class__).with_id(document.id)

        if new_instance:
            return new
        else:
            document.__data__ = new.__data__
            document.__fields_changed__.clear()
            return document

    def bulk(self, document_class, ordered=False, raise_on_errors=True):
        """ Return Bulk

        :param MetaDocument document_class: class of documents fo bulk
        :param bool ordered: create ordered bulk (default `False`)
        :param bool raise_on_errors: raise BulkWriteError exception
            if write errors (default `True`)

        Context manager:

            with db.bulk(Doc) as bulk:
                bulk.insert(doc_1)
                bulk.insert(doc_2)
        """
        return Bulk(self, document_class, ordered, raise_on_errors)



"""
Functions for serialize and deserialize data.
"""

from yadm.markers import NotLoaded


def to_mongo(document, exclude=(), include=None):
    """ Serialize document to MongoDB data

    :param BaseDocument document: document for serializing
    :param list exclude: exclude fields
    :param list include: include only fields (all by default)
    """
    result = {}

    for name, field in document.__fields__.items():
        if name in exclude:
            continue

        if include is not None and name not in include:
            continue

        if not hasattr(document, name):
            continue

        value = getattr(document, name)
        result[name] = field.to_mongo(document, value)

    if include:
        include_groups = {}

        for name in (f for f in include if '.' in f):
            first, last = name.split('.', 1)
            include_groups.setdefault(first, set()).add(last)

        for name, subinclude in include_groups.items():
            edoc = getattr(document, name)
            result[name] = to_mongo(edoc, include=subinclude)

    return result


def from_mongo(document_class, data, clear_fields_changed=True):
    """ Deserialize MongoDB data to document

    :param document_class: document class
    :param dict data: data from MongoDB
    :param bool clear_fields_changed: clear changed flags
        for new document (default True)
    """
    document = document_class(__initialized__=False)

    for name, field in document.__fields__.items():
        if name in data:
            value = data[name]
            document.__data__[name] = value

        else:
            document.__data__[name] = NotLoaded

    if clear_fields_changed:
        document.__fields_changed__.clear()

    document.__initialized__ = True
    return document



from bson import ObjectId

from yadm.documents import Document, EmbeddedDocument
from yadm import fields

from .test_database import BaseDatabaseTest


class ReferenceTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDocRef(Document):
            __collection__ = 'testdocs_ref'

        class TestDoc(Document):
            __collection__ = 'testdocs'
            ref = fields.ReferenceField(TestDocRef)

        self.TestDocRef = TestDocRef
        self.TestDoc = TestDoc

    def test_get(self):
        id_ref = self.db.db.testdocs_ref.insert({})
        id = self.db.db.testdocs.insert({'ref': id_ref})

        doc = self.db.get_queryset(self.TestDoc).with_id(id)

        self.assertEqual(doc._id, id)
        self.assertIsInstance(doc.ref, Document)
        self.assertEqual(doc.ref._id, id_ref)

    def test_set_objectid(self):
        id = self.db.db.testdocs.insert({})
        doc = self.db.get_queryset(self.TestDoc).with_id(id)

        doc.ref = self.db.db.testdocs_ref.insert({})
        self.db.save(doc)

        id_ref = self.db.db.testdocs.find_one({'_id': id})['ref']

        self.assertIsInstance(id_ref, ObjectId)
        self.assertIsInstance(doc.ref, Document)
        self.assertEqual(id_ref, doc.ref.id)

    def test_set_doc(self):
        id = self.db.db.testdocs.insert({})
        doc = self.db.get_queryset(self.TestDoc).with_id(id)

        doc_ref = self.TestDocRef()
        self.db.save(doc_ref)

        doc.ref = doc_ref
        self.db.save(doc)

        id_ref = self.db.db.testdocs.find_one({'_id': id})['ref']

        self.assertIsInstance(id_ref, ObjectId)
        self.assertIsInstance(doc.ref, Document)
        self.assertEqual(id_ref, doc.ref.id)


class ReferenceNotObjectIdTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDocRef(Document):
            __collection__ = 'testdocs_ref'
            _id = fields.IntegerField()

        class TestDoc(Document):
            __collection__ = 'testdocs'
            ref = fields.ReferenceField(TestDocRef)

        self.TestDocRef = TestDocRef
        self.TestDoc = TestDoc

    def test_get(self):
        id_ref = self.db.db.testdocs_ref.insert({'_id': 13})
        id = self.db.db.testdocs.insert({'ref': id_ref})

        doc = self.db.get_queryset(self.TestDoc).with_id(id)

        self.assertEqual(doc._id, id)
        self.assertIsInstance(doc.ref, Document)
        self.assertEqual(doc.ref._id, id_ref)


class ReferenceInEmbeddedTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class TestDocRef(Document):
            __collection__ = 'testdocs_ref'

        class TestDocEmb(EmbeddedDocument):
            ref = fields.ReferenceField(TestDocRef)

        class TestDoc(Document):
            __collection__ = 'testdocs'
            emb = fields.EmbeddedDocumentField(TestDocEmb)

        self.TestDocRef = TestDocRef
        self.TestDocEmb = TestDocEmb
        self.TestDoc = TestDoc

    def test_get(self):
        id_ref = self.db.db.testdocs_ref.insert({})
        id = self.db.db.testdocs.insert({'emb': {'ref': id_ref}})

        doc = self.db.get_queryset(self.TestDoc).with_id(id)

        self.assertEqual(doc._id, id)
        self.assertIsInstance(doc.emb, self.TestDocEmb)
        self.assertIsInstance(doc.emb.ref, self.TestDocRef)
        self.assertEqual(doc.emb.ref._id, id_ref)



from unittest import SkipTest

from pymongo.errors import BulkWriteError

from yadm import fields
from yadm.documents import Document
from yadm.bulk import Bulk

from .test_database import BaseDatabaseTest


class BulkTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        if self.db.db.connection.server_info()['version'] < '2.6':
            self.__class__.skip = True
            raise SkipTest('Bulk not work with MongoDB < 2.6')

        class TestDoc(Document):
            __collection__ = 'testdocs'

            i = fields.IntegerField()

        self.TestDoc = TestDoc

        self.db.db.testdocs.ensure_index([('i', 1)], unique=True)

    def test_create(self):
        bulk = self.db.bulk(self.TestDoc)
        self.assertIsInstance(bulk, Bulk)

    def test_insert_one(self):
        doc = self.TestDoc()
        doc.i = 1

        bulk = self.db.bulk(self.TestDoc)
        bulk.insert(doc)

        self.assertEqual(self.db.db.testdocs.count(), 0)

        bulk.execute()

        self.assertTrue(bulk.result)
        self.assertEqual(self.db.db.testdocs.count(), 1)
        self.assertEqual(self.db.db.testdocs.find_one()['i'], 1)

    def test_insert_many(self):
        bulk = self.db.bulk(self.TestDoc)

        for i in range(10):
            doc = self.TestDoc()
            doc.i = i
            bulk.insert(doc)

        self.assertEqual(self.db.db.testdocs.count(), 0)
        bulk.execute()
        self.assertEqual(self.db.db.testdocs.count(), 10)

    def test_insert_type_error(self):
        class Doc(Document):
            pass

        bulk = self.db.bulk(self.TestDoc)

        with self.assertRaises(TypeError):
            bulk.insert(Doc())

    def test_context_manager(self):
        with self.db.bulk(self.TestDoc) as bulk:
            doc = self.TestDoc()
            doc.i = 1
            bulk.insert(doc)

        self.assertEqual(self.db.db.testdocs.count(), 1)

    def test_context_manager_error(self):
        with self.assertRaises(RuntimeError):
            with self.db.bulk(self.TestDoc) as bulk:
                doc = self.TestDoc()
                doc.i = 1
                bulk.insert(doc)

                raise RuntimeError

        self.assertEqual(self.db.db.testdocs.count(), 0)

    def test_result(self):
        with self.db.bulk(self.TestDoc) as bulk:
            doc = self.TestDoc()
            doc.i = 1
            bulk.insert(doc)

        self.assertEqual(bulk.result.n_inserted, 1)

    def test_result_write_error(self):
        with self.db.bulk(self.TestDoc, raise_on_errors=False) as bulk:
            doc = self.TestDoc()
            doc.i = 1
            bulk.insert(doc)

            doc = self.TestDoc()
            doc.i = 2
            bulk.insert(doc)

            doc = self.TestDoc()
            doc.i = 1
            bulk.insert(doc)

        self.assertFalse(bulk.result)
        self.assertEqual(self.db.db.testdocs.count(), 2)
        self.assertEqual(bulk.result.n_inserted, 2)
        self.assertEqual(len(bulk.result.write_errors), 1)
        self.assertEqual(bulk.result.write_errors[0].document.i, 1)

    def test_result_write_error_raise(self):
        with self.assertRaises(BulkWriteError):
            with self.db.bulk(self.TestDoc) as bulk:
                doc = self.TestDoc()
                doc.i = 1
                bulk.insert(doc)

                doc = self.TestDoc()
                doc.i = 2
                bulk.insert(doc)

                doc = self.TestDoc()
                doc.i = 1
                bulk.insert(doc)

        self.assertEqual(bulk.result.n_inserted, 2)
        self.assertEqual(len(bulk.result.write_errors), 1)
        self.assertEqual(bulk.result.write_errors[0].document.i, 1)

    def test_result_write_error_ordered(self):
        with self.db.bulk(self.TestDoc, ordered=True, raise_on_errors=False) as bulk:
            for i in range(10):
                doc = self.TestDoc()
                doc.i = i
                bulk.insert(doc)

                doc = self.TestDoc()
                doc.i = 1
                bulk.insert(doc)

        self.assertEqual(self.db.db.testdocs.count(), 2)
        self.assertEqual(bulk.result.n_inserted, 2)
        self.assertEqual(len(bulk.result.write_errors), 1)
        self.assertEqual(bulk.result.write_errors[0].document.i, 1)



import sys
sys.path.insert(0, '.')

project = u'yadm'
copyright = u'2014, Alexander "ZZZ" Zelenyak'
version = '0.4'
release = version

extensions = ['sphinx.ext.autodoc']
templates_path = ['_templates']
source_suffix = '.rst'
source_encoding = 'utf8'
master_doc = 'index'
exclude_trees = ['_build']

# pygments_style = 'sphinx'
# html_theme = 'sphinxdoc'
# html_static_path = ['_static']



"""
Map

.. code-block:: python

    class Doc(Document):
        __collection__ = 'docs'
        map = fields.MapField(fields.IntegerField)

    doc = Doc()
    doc.map['a'] = 1
    doc.map['b'] = 2
    assert doc.map == {'a': 1, 'b': 2}

    db.insert(doc)
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload

    doc.map['c'] = 3  # do not save
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    del doc.map['b']  # do not save too
    assert doc.map == {'a': 1}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

    doc.map.set('d', 3)  # $set query
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2, 'c': 3}

    doc.map.unset('d', 3)  # $unset query
    assert doc.map == {'a': 1, 'b': 2}
    doc = db.get_queryset(Doc).with_id(doc.id)  # reload
    assert doc.map == {'a': 1, 'b': 2}

"""
from collections import abc

from bson import ObjectId

from yadm.fields.containers import (
    Container,
    ContainerField,
)


class Map(Container, abc.MutableMapping):
    """ Map
    """
    def __iter__(self):
        return iter(self._data)

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self._data)

    def _load_from_mongo(self, data):
        self._data = {}

        for key, value in (data or {}).items():
            if hasattr(self._field.value_field, 'from_mongo'):
                value = self._field.value_field.from_mongo(self.__parent__, value)
            else:
                value = self._prepare_value(value)

            self._data[key] = value

    def _prepare_value(self, value):
        """ `prepare_value` function for `value_field`
        """
        if hasattr(self._field.value_field, 'prepare_value'):
            return self._field.value_field.prepare_value(self.__parent__, value)
        else:
            return value

    def set(self, key, value):
        """ Set key directly in database

        :param key: key
        :param value: value for `$set`

        See `$set` in MongoDB's `set`.
        """
        key = str(key)

        if hasattr(self._field.value_field, 'to_mongo'):
            value = self._field.value_field.to_mongo(self.__document__, value)
        else:
            value = self._prepare_value(value)

        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$set': {fn: value}}, multi=False)
        self._data[key] = value

    def unset(self, key):
        """ Unset key directly in database

        :param key: key

        See `$unset` in MongoDB's `unset`.
        """
        key = str(key)
        qs = self._get_queryset()
        fn = '.'.join([self.__field_name__, key])
        qs.update({'$unset': {fn: True}}, multi=False)
        del self._data[key]


class MapField(ContainerField):
    """ Field for maps
    """
    container = Map

    def __init__(self, value_field):
        self.value_field = value_field

    @property
    def default(self):
        return {}

    def from_mongo(self, document, value):
        if isinstance(value, self.container):
            return value
        else:
            return self.container(document, self, value)

    def to_mongo(self, document, value):
        tm = self.value_field.to_mongo
        return {k: tm(document, v) for k, v in value._data.items()}


class MapCustomKeys(Map):
    def from_str(self, value):
        """ Cast value """
        raise NotImplementedError

    def _load_from_mongo(self, data):
        data = {str(k): v for k, v in data.items()}
        super()._load_from_mongo(data)

    def __iter__(self):
        from_str = self.from_str
        return (from_str(k) for k in super().__iter__())

    def __getitem__(self, key):
        return super().__getitem__(str(key))

    def __setitem__(self, key, value):
        super().__setitem__(str(key), value)

    def __delitem__(self, key):
        super().__delitem__(str(key))

    def __contains__(self, key):
        return str(key) in self._data


class MapCustomKeysField(MapField):
    container = MapCustomKeys

    def prepare_value(self, document, value):
        if value is None:
            return None
        elif isinstance(value, self.container):
            return value
        else:
            value = {str(k): v for k, v in value.items()}
            return self.container(document, self, value)


class MapIntKeys(MapCustomKeys):
    from_str = int


class MapIntKeysField(MapCustomKeysField):
    container = MapIntKeys


class MapObjectIdKeys(MapCustomKeys):
    from_str = ObjectId


class MapObjectIdKeysField(MapCustomKeysField):
    container = MapObjectIdKeys



from bson import ObjectId

from yadm import fields
from yadm.documents import Document, EmbeddedDocument

from .test_database import BaseDatabaseTest


class EmbeddedDocumentFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class ETestDoc(EmbeddedDocument):
            i = fields.IntegerField()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            e = fields.EmbeddedDocumentField(ETestDoc)

        self.ETestDoc = ETestDoc
        self.TestDoc = TestDoc

    def test_default(self):
        td = self.TestDoc()
        self.assertFalse(hasattr(td, 'e'))

    def test_get(self):
        _id = self.db.db.testdoc.insert({'e': {'i': 13}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)

        self.assertTrue(hasattr(td, 'e'))
        self.assertIsInstance(td.e, self.ETestDoc)
        self.assertTrue(hasattr(td.e, 'i'))
        self.assertIsInstance(td.e.i, int)
        self.assertEqual(td.e.i, 13)

    def test_set(self):
        td = self.TestDoc()
        td.e = self.ETestDoc()

        self.assertTrue(hasattr(td, 'e'))
        self.assertIsInstance(td.e, self.ETestDoc)
        self.assertFalse(hasattr(td.e, 'i'))

        td.e.i = 13
        self.assertTrue(hasattr(td.e, 'i'))
        self.assertIsInstance(td.e.i, int)
        self.assertEqual(td.e.i, 13)

    def test_set_typeerror(self):
        class FailETestDoc(EmbeddedDocument):
            s = fields.StringField

        td = self.TestDoc()
        self.assertRaises(TypeError, setattr, td, 'e', FailETestDoc())

    def test_set_insert(self):
        td = self.TestDoc()
        td.e = self.ETestDoc()
        td.e.i = 13
        self.db.insert(td)

        data = self.db.db.testdoc.find_one({'_id': td.id})
        self.assertEqual(data, {'_id': td.id, 'e': {'i': 13}})

    def test_set_save(self):
        _id = self.db.db.testdoc.insert({'e': {'i': 13}})
        td = self.db.get_queryset(self.TestDoc).with_id(_id)

        td.e.i = 26
        # td.__fields_changed__.add('e.i')
        self.db.save(td)

        data = self.db.db.testdoc.find_one({'_id': _id})
        self.assertEqual(data, {'_id': _id, 'e': {'i': 26}})


class EmbeddedDocumentWithIdFieldTest(BaseDatabaseTest):
    def setUp(self):
        super().setUp()

        class ETestDoc(EmbeddedDocument):
            id = fields.ObjectIdField(default_gen=True)
            i = fields.IntegerField()

        class TestDoc(Document):
            __collection__ = 'testdoc'
            e = fields.EmbeddedDocumentField(ETestDoc)

        self.ETestDoc = ETestDoc
        self.TestDoc = TestDoc

    def test_load_default(self):
        td = self.TestDoc({'e': {'i': 13}})
        self.assertTrue(hasattr(td, 'e'))
        self.assertTrue(hasattr(td.e, 'i'))
        self.assertTrue(hasattr(td.e, 'id'))
        self.assertEqual(td.e.i, 13)
        self.assertIsInstance(td.e.id, ObjectId)
